#pragma once

#include <QAbstractItemModel>
#include <QDebug>

#include <memory>

template<class Data>
struct TreeItem
{
    TreeItem() = default;
    TreeItem(Data data, TreeItem* parent = nullptr)
        : parent{parent}
        , data{std::move(data)}
    {
    }

    int             calcRow() const;
    int             childCount() const { return static_cast<int>(children.size()); };
    const TreeItem* child(int index) const;

    TreeItem* appendChild(Data childData);
    TreeItem* appendChild(std::unique_ptr<TreeItem> child);

    void reset() { children.clear(); }

    TreeItem* parent = {};
    Data      data;
    std::vector<std::unique_ptr<TreeItem>> children;
};

template<class Data>
TreeItem<Data>* TreeItem<Data>::appendChild(Data childData)
{
    auto child = std::make_unique<TreeItem>(childData, this);
    return children.emplace_back(std::move(child)).get();
}

template<class Data>
TreeItem<Data>* TreeItem<Data>::appendChild(std::unique_ptr<TreeItem<Data>> child)
{
    child->parent = this;
    return children.emplace_back(std::move(child)).get();
}

template<class Data>
int TreeItem<Data>::calcRow() const
{
    auto isThis = [this](const std::unique_ptr<TreeItem>& item){ return item.get() == this; };

    if (parent)
    {
        if (auto it = std::find_if(parent->children.cbegin(), parent->children.cend(), isThis);
                it != parent->children.cend())
        {
            return std::distance(parent->children.cbegin(), it);
        } else {
            qWarning() << "Tree child item not found in parent";
            return 0;
        }
    }

    return 0;
}

template<class Data>
const TreeItem<Data> *TreeItem<Data>::child(int index) const
{
    if (index >= 0 && static_cast<size_t>(index) < children.size())
        return children[index].get();
    else
        return {};
}

template<class Data>
TreeItem<Data>* appendItem(TreeItem<Data> &parent, std::unique_ptr<TreeItem<Data>> data)
{
    parent.children.push_back(std::move(data));
    parent.children.back()->parent = &parent;

    return parent.children.back().get();
}

/**
    Implements a tree data structure that conforms the QAbstractItemModel API and can be used with Qt
    tree views.

    Data is a class for user-defined data stored in a tree row.

    *** Important info for implementors***
    - the column count of a tree model must be fixed! You cannot add/remove columns on the fly. Use a filter proxy
      model to e.g. hide/show/reorder columns.
*/
template<class Data>
class TreeModel : public QAbstractItemModel
{
public:
    using DataType = Data;
    using TreeItemType = TreeItem<DataType>;

    explicit TreeModel(QObject* parent = nullptr);

    virtual QModelIndex index(int row, int column, const QModelIndex &parent) const override;
    virtual QModelIndex parent(const QModelIndex &child) const override;
    virtual int rowCount(const QModelIndex &parent) const override;

          TreeItemType *getTreeItem(const QModelIndex& index);
    const TreeItemType *getTreeItem(const QModelIndex& index) const;
    const TreeItemType *getTreeItemOrRoot(const QModelIndex& index) const;

    /**
        Creates an index for \a treeItem. This will calculate the row the tree item is at, if you already know,
        use the overload that takes a row argument
    */
    QModelIndex createItemIndex(const TreeItemType* treeItem) const;

    /**
        Create an index for a tree item where you know the row index it has in its parent
    */
    QModelIndex createItemIndex(int row, int column, const TreeItemType* parentItem) const;

    /**
        Sets the parent of \data to \a parent
    */
    TreeItemType *appendItem(TreeItemType &parent, std::unique_ptr<TreeItemType> data);
    void            reset();

    TreeItemType *root() { return &m_root; }
    TreeItemType const *root() const { return &m_root; }
    /**
        The return value of `updater` should indicate whether an update really took place and hence
        UI notifications should be fired
    */
    void            update(const QModelIndex& index, std::function<bool (Data&)> updater,
                           const QVector<int>& roles);

    QModelIndex     rootIndex() const;

private:
    TreeItemType m_root;
};

template<class Data>
TreeModel<Data>::TreeModel(QObject* parent /* = nullptr */)
    : QAbstractItemModel(parent)
{
}

template<class Data>
auto TreeModel<Data>::appendItem(TreeItemType &parent, std::unique_ptr<TreeItemType> data) -> TreeItemType*
{
    beginInsertRows(createItemIndex(&parent), parent.childCount(), parent.childCount());
    auto pItem = parent.appendChild(std::move(data));
    endInsertRows();

    return pItem;
}

template<class Data>
void TreeModel<Data>::update(const QModelIndex& index, std::function<bool (Data&)> updater,
                             const QVector<int>& roles)
{
    if (auto item = getTreeItem(index))
    {
        if (updater(item->data))
            dataChanged(index, index, roles);
    }
}

template<class Data>
void TreeModel<Data>::reset()
{
    beginResetModel();
    m_root.reset();
    endResetModel();
}

template<class Data>
QModelIndex TreeModel<Data>::rootIndex() const
{
    return {};
}

template<class Data>
QModelIndex TreeModel<Data>::index(int row, int column, const QModelIndex &parent) const
{
    if (parent.isValid() && parent.column() != 0)
        return QModelIndex();

    if (auto parentItem = getTreeItemOrRoot(parent))
        return createItemIndex(row, column, parentItem->child(row));

    return {};
}

template<class Data>
QModelIndex TreeModel<Data>::parent(const QModelIndex &child) const
{
    if (!child.isValid())
        return {};

    const TreeItemType *childItem = getTreeItem(child);
    // Second case *should* never happen, catch it anyway
    if (childItem->parent == root() || !childItem->parent)
        return {};

    return createItemIndex(childItem->parent->calcRow(), 0, childItem->parent);
}

template<class Data>
int TreeModel<Data>::rowCount(const QModelIndex &parent) const
{
    if (parent.column() > 0)
        return 0;

    if (const TreeItemType *parentItem = getTreeItemOrRoot(parent))
        return static_cast<int>(parentItem->children.size());
    return 0;
}

template<class Data>
auto TreeModel<Data>::getTreeItem(const QModelIndex &index) -> TreeItemType*
{
    return const_cast<TreeItemType*>(static_cast<const TreeModel*>(this)->getTreeItem(index));
}

template<class Data>
auto TreeModel<Data>::getTreeItem(const QModelIndex &index) const -> const TreeItemType*
{
    if (index.isValid())
        return static_cast<TreeItemType*>(index.internalPointer());
    return {};
}

template<class Data>
auto TreeModel<Data>::getTreeItemOrRoot(const QModelIndex &index) const -> const TreeItemType*
{
    if (!index.isValid())
        return root();
    return static_cast<TreeItemType*>(index.internalPointer());
}

template<class Data>
QModelIndex TreeModel<Data>::createItemIndex(const TreeItemType *treeItem) const
{
    if (treeItem)
    {
        return createItemIndex(treeItem->calcRow(), 0, treeItem);
    }
    return {};
}

template<class Data>
QModelIndex TreeModel<Data>::createItemIndex(int row, int column, const TreeItemType *parentItem) const
{
    // The root item has an invalid model index, that's why we have != m_root here
    if (parentItem && parentItem != root())
        return createIndex(row, column, const_cast<TreeItemType*>(parentItem));
    return {};
}
